/*
 * yasapi.h
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#ifndef __YASAPI_H__ // {
#define __YASAPI_H__
#include <windows.h>
#include <shlobj.h>
#include <audioclient.h>
#include <mmdeviceapi.h>
#include <mftransform.h>
#include <mferror.h>
#include <mfapi.h>
#define tagTIMECODE_SAMPLE ____tagTIMECODE_SAMPLE
#define TIMECODE_SAMPLE ____TIMECODE_SAMPLE
#define PTIMECODE_SAMPLE ____PTIMECODE_SAMPLE
#define _DDPIXELFORMAT _____DDPIXELFORMAT
#define DDPIXELFORMAT ____DDPIXELFORMAT
#define LPDDPIXELFORMAT ____LPDDPIXELFORMAT
#include <wmcodecdsp.h>
#undef LPDDPIXELFORMAT
#undef DDPIXELFORMAT
#undef _DDPIXELFORMAT
#undef PTIMECODE_SAMPLE
#undef TIMECODE_SAMPLE
#undef tagTIMECODE_SAMPLE
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include <winamp/wa_ipc.h>
#include <winamp/out.h>
#ifdef __cpluplus
extern "C" {
#endif

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_CONFIG2
//#define YASAPI_OPTION_WRITTEN_TIME

#if defined (_MSC_VER) && ! defined (__func__) // {
#define __func__ __FUNCTION__
#endif // }

// http://stackoverflow.com/questions/2192416/how-to-convert-concatenated-strings-to-wide-char-with-the-c-preprocessor
#define YASAPI_STR1(x)    #x
#define YASAPI_STR(x)     YASAPI_STR1(x)

#define YASAPI_WSTR1(x)   L ## #x
#define YASAPI_WSTR(x)    YASAPI_WSTR1(x)

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPID) // {
  #define YASAPI_DEBUG
  #if ! defined (YASAPI_ID) // {
  	#define YASAPI_ID out_yasapid
  #endif // }
#else // } {
  //#define YASAPI_DEBUG
  #if ! defined (YASAPI_ID) // {
  	#define YASAPI_ID out_yasapi
  #endif // }
#endif // }

#if ! defined (YASAPI_VER) // {
  #define YASAPI_VER 0.0.0
#endif // }

#define YASAPI_NAME       YASAPI_STR(YASAPI_ID)
#define YASAPI_VERSION    YASAPI_STR(YASAPI_VER)
#define YASAPI_MODULE     YASAPI_NAME ".dll"

#define YASAPI_INT_MAX(n) \
    (~(~0<<((n)-1)))
#define YASAPI_SHAREMODE_LABEL(eShareMode) \
    (AUDCLNT_SHAREMODE_SHARED==(eShareMode)?"SHARED":"EXCLUSIVE")

#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
#define YASAPI_MODE_LABEL(mode) ( \
    YASAPI_MODE_SHARED==(mode)?"SHARED": \
    YASAPI_MODE_EXCLUSIVE==(mode)?"EXCLUSIVE": \
    "AUTOMATIC" \
)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#define PI_VER2 "v" YASAPI_VERSION

#if defined (YASAPID) // {
#define PI_VER PI_VER2 " (debug)"
#else // } {
#define PI_VER PI_VER2
#endif // }

#define PI_LABEL "Yet Another (WA)SAPI Output " PI_VER

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEBUG) // {
  #define DPUTS(debug,cs) \
      ((debug)&&fputs(cs,stderr))
  #define DPRINTF(debug,cs,...) \
      ((debug)&&fprintf(stderr,cs,__VA_ARGS__))
  #define DPUTWS(debug,ws) \
      ((debug)&&fputws(ws,stderr))
  #define DWPRINTF(debug,ws,...) \
      ((debug)&&fwprintf(stderr,ws,__VA_ARGS__))
  #define DERROR(debug,x,y) \
      ((debug)&&(x)==(y)&&fputs(#x "\n",stderr))
  #define DMESSAGE(debug,m) (debug&&fprintf(stderr,"Error " m ": " \
      "\"%s\" (%d).\n",basename(__FILE__),__LINE__))
  #define DMESSAGEV(debug,m,...) (debug&&fprintf(stderr,"Error " m ": " \
      "\"%s\" (%d).\n",__VA_ARGS__,basename(__FILE__),__LINE__))
#else // } {
  #define DPUTS(debug,cs)
  #define DPRINTF(debug,cs,...)
  #define DPUTWS(debug,ws)
  #define DWPRINTF(debug,ws,...)
  #define DERROR(debug,x,y) \
    ((x)==(y)&&message(#x ": \"%s\" (%d).",basename(__FILE__),__LINE__))
  #define DMESSAGE(debug,m)
  #define DMESSAGEV(debug,m,...)
#endif // }

///////////////////////////////////////////////////////////////////////////////
DEFINE_GUID(IID_IUnknown, 0x00000000, 0x0000, 0x0000, 0xc0,0x00,
    0x00,0x00,0x00,0x00,0x00,0x46);

DEFINE_GUID(IID_IAudioClient, 0x1cb9ad4c, 0xdbfa, 0x4c32, 0xb1,0x78,
    0xc2,0xf5,0x68,0xa7,0x03,0xb2);
DEFINE_GUID(IID_IAudioClock, 0xcd63314f, 0x3fba, 0x4a1b, 0x81,0x2c,
    0xef,0x96,0x35,0x87,0x28,0xe7);
DEFINE_GUID(IID_IAudioRenderClient, 0xf294acfc, 0x3146, 0x4483, 0xa7,0xbf,
    0xad,0xdc,0xa7,0xc2,0x60,0xe2);

DEFINE_GUID(CLSID_MMDeviceEnumerator, 0xbcde0395, 0xe52f, 0x467c, 0x8e,0x3d,
    0xc4,0x57,0x92,0x91,0x69,0x2e);
DEFINE_GUID(IID_IMMDeviceEnumerator, 0xa95664d2, 0x9614, 0x4f35, 0xa7,0x46,
    0xde,0x8d,0xb6,0x36,0x17,0xe6);
DEFINE_GUID(IID_IMMDevice, 0xd666063f, 0x1587, 0x4e43, 0x81,0xf1,
    0xb9,0x48,0xe8,0x07,0x36,0x3f);

#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
///////////////////////////////////////////////////////////////////////////////
#define YASAPI_MODE_SHARED      0
#define YASAPI_MODE_EXCLUSIVE   1
#define YASAPI_MODE_AUTOMATIC   2
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_CONFIG2) // {
#define YASAPI_ID_SIZE 128
#endif // }

typedef struct SyncConfig SyncConfig;
typedef struct SyncChannel SyncChannel;
typedef struct Sync Sync;
typedef struct Timer Timer;
typedef struct Ring Ring;
typedef struct Plugin Plugin;
typedef struct PlayerStrategy PlayerStrategy;
typedef struct Player Player;
#if defined (YASAPI_CONFIG2) // {
typedef wchar_t yasapi_id_t[YASAPI_ID_SIZE];
typedef struct DeviceOptions DeviceOptions;
#endif // }
typedef struct Options Options;
typedef struct Slider Slider;

///////////////////////////////////////////////////////////////////////////////
__declspec( dllexport ) Out_Module * winampGetOutModule();

char *basename(char *s);
#if ! defined (YASAPI_DEBUG) // {
int message(const char *format, ...);
#endif // }

void AboutDialog(HWND hWndParent);
void ConfigDialog(HWND hWndParent, Options *pOptions);
#if defined (YASAPI_DEBUG) // {
void ConsoleOpen(void);
void ConsoleClose(void);
#endif // }

///////////////////////////////////////////////////////////////////////////////
void WFXXSetup(WAVEFORMATEXTENSIBLE *pwfxx, int samplerate, int numchannels,
    int bitspersamp, BOOL bSurround, BOOL bFloat);
#if defined (YASAPI_DEBUG) // {
const char *WFXXGetSubFormat(const WAVEFORMATEXTENSIBLE *pwfxx);
void WFXPrint(const WAVEFORMATEX *pwfx, FILE *f);
#endif // }

///////////////////////////////////////////////////////////////////////////////
HRESULT MediaTypeCreateWFX(IMFMediaType **ppMediaType, WAVEFORMATEX *pwfx,
    const char *name, int debug);

///////////////////////////////////////////////////////////////////////////////
HRESULT TransformProcessMessage(IMFTransform *pTransform, ULONG_PTR ulParam,
    const char *name, int debug);

///////////////////////////////////////////////////////////////////////////////
#define SYNC_IS(pSync,flags) \
    (0!=((pSync)->state&(flags)))

typedef BOOL (*SyncCondition)(LPVOID data);

struct SyncConfig {
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }
  LPVOID pClient;

  struct {
    int state;
  } end;

  struct {
    int state;
    SyncCondition Condition;
  } read;

  struct {
    int state;
    SyncCondition Condition;
  } write;
};

struct SyncChannel {
  int state;
  SyncCondition Condition;
  HANDLE hEvent;
};

struct Sync {
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }
  volatile int state;
  LPVOID pClient;
  HANDLE hMutex;

  struct {
    int state;
  } end;

  SyncChannel read;
  SyncChannel write;
};

DWORD SyncCreate(Sync *pSync, SyncConfig *c);
void SyncDestroy(Sync *pSync);

void SyncWaitRead(Sync *pSync);
void SyncReleaseRead(Sync *pSync);

void SyncWaitWrite(Sync *pSync);
void SyncReleaseWrite(Sync *pSync);

///////////////////////////////////////////////////////////////////////////////
struct Timer {
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }
  HANDLE hTimer;
};

int TimerCreate(Timer *pTimer, int debug);
void TimerDestroy(Timer *pTimer);

void TimerKill(Timer *pTimer);
void TimerSleep(Timer *pTimer, LONGLONG time);
void TimerSleepTime(Timer *pTimer, LONGLONG time);
void TimerSleepMS(Timer *pTimer, LONGLONG ms);

///////////////////////////////////////////////////////////////////////////////
typedef void (*RingCopy)(PVOID pClient, PVOID Destination,
    const VOID *Source, SIZE_T Length);

struct Ring {
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }
  LPSTR pRep;
  LPSTR pMax;
  volatile LPSTR pHead;
  volatile LPSTR pTail;
  volatile SIZE_T dwWritten;
  volatile SIZE_T dwAvailable;
  RingCopy Copy;
  PVOID pClient;
};

DWORD RingCreate(Ring *pRing, SIZE_T dwSize, int debug);
void RingDestroy(Ring *pRing);

void RingReset(Ring *pRing);
SIZE_T RingWrite(Ring *pRing, LPSTR pData, SIZE_T dwSize, SIZE_T n);
SIZE_T RingRead(Ring *pRing, LPSTR pData, SIZE_T dwSize);

void RingCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length);

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_PULL

#define PLAYER_IS_WRITE(pSync)    SYNC_IS(pSync,PLAYER_WRITE)
#define PLAYER_IS_READ(pSync)     SYNC_IS(pSync,PLAYER_READ)
#define PLAYER_IS_STARTED(pSync)  SYNC_IS(pSync,PLAYER_STARTED)
#define PLAYER_IS_PAUSE(pSync)    SYNC_IS(pSync,PLAYER_PAUSE)
#define PLAYER_IS_END(pSync)      SYNC_IS(pSync,PLAYER_END)

enum {
  PLAYER_WRITE=1<<1,
  PLAYER_READ=1<<2,
  PLAYER_STARTED=1<<3,
  PLAYER_PAUSE=1<<4,
  PLAYER_END=1<<5,
};

struct Player {
  int initialized;
  const PlayerStrategy *strategy;
  Plugin *pPlugin;
  WAVEFORMATEXTENSIBLE wfxx;
  WORD wBytesPerSample;
  int mono;
  IAudioClient *pAudioClient;
  HANDLE hEvent;
  IAudioRenderClient *pRenderClient;
  UINT32 nFrames;
  Sync sync;
  Ring ring;
  HANDLE hThread;
  IAudioClock *pAudioClock;
};

int PlayerCreate(Player *pPlayer, Plugin *pPlugin, int samplerate,
    int numchannels, int bitspersamp);
void PlayerDestroy(Player *pPlayer);

void PlayerShutdown(Player *pPlayer);
#if defined (YASAPI_DEBUG) // {
int PlayerGetDebug(Player *pPlayer);
#endif // }
int PlayerGetMaxLatencyMs(Player *pPlayer);

BOOL PlayerCanWrite(LPVOID pClient);
BOOL PlayerCanRead(LPVOID pClient);

HRESULT PlayerStart(Player *pPlayer);
HRESULT PlayerStopUnlocked(Player *pPlayer);
HRESULT PlayerStop(Player *pPlayer);
DWORD PlayerGetTime(Player *pPlayer);
WORD PlayerSamplesPerSec(Player *pPlayer);
WORD PlayerChannels(Player *pPlayer);
WORD PlayerBlockAlign(Player *pPlayer);
SIZE_T PlayerAvailable(Player *pPlayer);
SIZE_T PlayerWriteWait(Player *pPlayer, LPSTR pData, SIZE_T dwSize);
SIZE_T PlayerWrite(Player *pPlayer, LPSTR pData, SIZE_T dwSize);
void PlayerReset(Player *pPlayer);
void PlayerPause(Player *pPlayer, BOOL bPause);

HRESULT PlayerRenderFrames(Player *pPlayer, UINT32 *pNumFramesRender);
DWORD WINAPI PlayerThread(LPVOID lpParameter);

void PlayerCopyMemory(PVOID pClient, PVOID Destination, const VOID *Source,
    SIZE_T Length);

///////////////////////////////////////////////////////////////////////////////
struct PlayerStrategy {
  const char *name;

  HRESULT (*InitializeAudioClient)(Player *pPlayer,
      AUDCLNT_SHAREMODE eShareMode, DWORD dwStreamFlags,
      REFERENCE_TIME hnsDevicePeriod);
  int (*CreateEvent)(Player *pPlayer);
  void (*DestroyEvent)(Player *pPlayer);

  BOOL (*CanRead)(Player *pPlayer);
  int (*Wait)(Player *pPlayer, LONGLONG time);
  HRESULT (*GetFramesWrite)(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding);
};

const PlayerStrategy *PlayerPush(void);
const PlayerStrategy *PlayerPull(void);

HRESULT PlayerGetFramesWritePush(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding);
HRESULT PlayerGetFramesWritePull(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding);

///////////////////////////////////////////////////////////////////////////////
struct Plugin {
  int initialized;
  CRITICAL_SECTION cs;
  Options *pOptions;
  double volume;
  Timer timer;
  IMMDeviceEnumerator *pEnumerator;
#if defined (YASAPI_CONFIG2) // {
  IMMDeviceCollection *pCollection;
#else // } {
  IMMDevice *pDevice;
#endif // }
};

int PluginCreate(Plugin *pPlugin, Options *pOptions);
void PluginDestroy(Plugin *pPlugin);

#if defined (YASAPI_CONFIG2) // {
///////////////////////////////////////////////////////////////////////////////
struct DeviceOptions {
  wchar_t *path;
  IMMDevice *pDevice;
  IPropertyStore *pProperties;

  yasapi_id_t id;
  PROPVARIANT val_Device_FriendlyName;
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  AUDCLNT_SHAREMODE eShareMode;
#else // } {
  int mode;
#endif // }
  int pull;
  int autoconvertpcm;
  int src_default_quality;
  double qRingSize;
  double qRingFill;
  double qShareSize;
  REFERENCE_TIME hnsDevicePeriod;
};

int DeviceOptionsInit(DeviceOptions *pOptions, Plugin *pPlugin, int nDevice);
void DeviceOptionsCleanup(DeviceOptions *pOptions);

const DeviceOptions *GetDefaultDeviceOptions(void);
void DeviceOptionsLoad(DeviceOptions *pOptions, wchar_t *app,
    const DeviceOptions *(*GetDefaultOptions)(void));
void DeviceOptionsSave(DeviceOptions *pOptions, wchar_t *app);
#else // } {
void DeviceOptionsLoad(Options *pOptions, wchar_t *app,
    const Options *(*GetDefaultOptions)(void));
void DeviceOptionsSave(Options *pOptions, wchar_t *app);
#endif // }

///////////////////////////////////////////////////////////////////////////////
struct Options {
  wchar_t *path;
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }
  int mono2stereo;
  int volume;

#if defined (YASAPI_OPTION_WRITTEN_TIME) // {
  struct {
    int audio;
    int video;
  } writtentime;
#endif // }

#if ! defined (YASAPI_CONFIG2) // {
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  AUDCLNT_SHAREMODE eShareMode;
#else // } {
  int mode;
#endif // }
  int pull;
  int autoconvertpcm;
  int src_default_quality;
  double qRingSize;
  double qRingFill;
  double qShareSize;
  REFERENCE_TIME hnsDevicePeriod;
#else // }
  yasapi_id_t id;
  int nPage;
  int nDevice;
  UINT cDevices;
  DeviceOptions *pDevices;
#endif // }
};

int OptionsCreate(Options *pOptions);
void OptionsDestroy(Options *pOptions);

#if defined (YASAPI_CONFIG2) // {
void OptionsSetDevice(Options *pOptions, int nDevice);
#endif // }
void OptionsLoad(Options *pOptions);
#if defined (YASAPI_DEBUG) // {
void OptionsSave(Options *pOptions, int debug);
#else // } {
void OptionsSave(Options *pOptions);
#endif // }

#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
AUDCLNT_SHAREMODE OptionsGetShareMode(const Options *pOptions);
#else // } {
AUDCLNT_SHAREMODE OptionsGetMode(const Options *pOptions);
#endif // }
int OptionsGetPull(const Options *pOptions);
int OptionsGetAutoConvertPCM(const Options *pOptions);
int OptionsGetSRCDefaultQuality(const Options *pOptions);
double OptionsGetRingSize(const Options *pOptions);
double OptionsGetRingFill(const Options *pOptions);
double OptionsGetShareSize(const Options *pOptions);

void OptionsSetDevicePeriod(const Options *pOptions,
    REFERENCE_TIME hnsDevicePeriod);
REFERENCE_TIME OptionsGetDevicePeriod(const Options *pOptions);

void OptionsCopy(Options *lhs, const Options *rhs);
const Options *GetDefaultOptions(void);

#if defined (YASAPI_DEBUG) // {
void OptionsDebug(Options *pOptions);
#endif // }

///////////////////////////////////////////////////////////////////////////////
#define CONFIG_SLIDER_SCALE   1000
#define CONFIG_SLIDER_MIN     (CONFIG_SLIDER_SCALE*1)
#define CONFIG_SLIDER_MAX     (CONFIG_SLIDER_SCALE*8)
#define CONFIG_SLIDER_DELTA   (CONFIG_SLIDER_MAX-CONFIG_SLIDER_MIN)

void ConfigSliderSetWidth(HWND hDlg, int nIdSlider, int nIdParent);
void ConfigSliderSetText(HWND hDlg, int nIdStatic, int nIdSlider);
double ConfigSliderGet(HWND hDlg, int nIdSlider);
void ConfigSliderSync(HWND hDlg);

void ConfigOptionsSyncShareMode(HWND hDlg);

#if defined (YASAPI_CONFIG2) // {
void ConfigOptionsSetDeviceOptions(HWND hDlg, const Options *pOptions,
    int nDevice);
void ConfigOptionsSetDeviceBuffers(HWND hDlg, const Options *pOptions,
    int nDevice);
#endif // }
void ConfigOptionsSetOptions(HWND hDlg, const Options *pOptions);

#if defined (YASAPI_CONFIG2) // {
void ConfigOptionsGetDeviceOptions(HWND hDlg, Options *pOptions, int nDevice);
void ConfigOptionsGetDeviceBuffers(HWND hDlg, Options *pOptions, int nDevice);
#endif // }
void ConfigOptionsGetOptions(HWND hDlg, Options *pOptions);

#ifdef __cpluplus
}
#endif
#endif // }
