/*
 * yasapi_config_common.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <resources.h>

///////////////////////////////////////////////////////////////////////////////
static void ConfigGetRect(HWND hDlg, int nId, RECT *rc)
{
  GetWindowRect(
    GetDlgItem(hDlg,nId),       // _In_   HWND hWnd,
    rc                          // _Out_  LPRECT lpRect
  );

  ScreenToClient(
    hDlg,                       // _In_  HWND hWnd,
    (LPPOINT)&rc->left          // LPPOINT lpPoint
  );

  ScreenToClient(
    hDlg,                       // _In_  HWND hWnd,
    (LPPOINT)&rc->right         // LPPOINT lpPoint
  );
}

void ConfigSliderSetWidth(HWND hDlg, int nIdSlider, int nIdParent)
{
  enum { MIN_WIDTH=27 };
  RECT rcSlider,rcParent;
  int pos,max,width;
  double q;

  // get the parent's maximum range.
  max=SendMessage(
    GetDlgItem(hDlg,nIdParent),
                                // _In_  HWND hWnd,
    TBM_GETRANGEMAX,            // _In_  UINT Msg,
    0,                          // _In_  WPARAM wParam,
    0                           // _In_  LPARAM lParam
  );

  // get the parent's position.
  pos=SendMessage(
    GetDlgItem(hDlg,nIdParent),
                                // _In_  HWND hWnd,
    TBM_GETPOS,                 // _In_  UINT Msg,
    0,                          // _In_  WPARAM wParam,
    0                           // _In_  LPARAM lParam
  );

  // calculate the scaling factor.
  q=pos-CONFIG_SLIDER_MIN;

  if (0.0<q)
    q/=(max-CONFIG_SLIDER_MIN);

  // calculate the slider's width based on the parent's width.
  ConfigGetRect(hDlg,nIdSlider,&rcSlider);
  ConfigGetRect(hDlg,nIdParent,&rcParent);

  width=MIN_WIDTH;
  width+=q*(rcParent.right-rcParent.left-MIN_WIDTH)+0.5;
  rcSlider.right=rcSlider.left+width;

  // resize the slider.
  MoveWindow(
    GetDlgItem(hDlg,nIdSlider),
                                // _In_  HWND hWnd,
    rcSlider.left,              // _In_  int X,
    rcSlider.top,               // _In_  int Y,
    width,                      // _In_  int nWidth,
    rcSlider.bottom-rcSlider.top,
                                // _In_  int nHeight,
    TRUE                        // _In_  BOOL bRepaint
  );

  // set the maximum to the parent's position.
  SendMessage(
    GetDlgItem(hDlg,nIdSlider),
                                // _In_  HWND hWnd,
    TBM_SETRANGE,               // _In_  UINT Msg,
    FALSE,                      // _In_  WPARAM wParam (redraw flag),
    MAKELONG(CONFIG_SLIDER_MIN,pos)
                                // _In_  LPARAM lParam
  );
}

void ConfigSliderSetText(HWND hDlg, int nIdStatic, int nIdSlider)
{
  double scale=ConfigSliderGet(hDlg,nIdSlider);
  char buf[128];

  sprintf(buf,"%.2f \xD7 minimum",scale);

  SetWindowTextA(
    GetDlgItem(hDlg,nIdStatic), // _In_      HWND hWnd,
    buf                         // _In_opt_  LPCTSTR lpString
  );
}

double ConfigSliderGet(HWND hDlg, int nIdSlider)
{
  double pos;

  pos=SendMessage(
    GetDlgItem(hDlg,nIdSlider),
                                // _In_  HWND hWnd,
    TBM_GETPOS,                 // _In_  UINT Msg,
    0,                          // _In_  WPARAM wParam,
    0                           // _In_  LPARAM lParam
  );

  return pos/CONFIG_SLIDER_SCALE;
}

void ConfigSliderSync(HWND hDlg)
{
  ConfigSliderSetText(hDlg,IDC_STATIC_RING_SIZE,IDC_SLIDER_RING_SIZE);

  ConfigSliderSetWidth(hDlg,IDC_SLIDER_RING_FILL,IDC_SLIDER_RING_SIZE);
  ConfigSliderSetText(hDlg,IDC_STATIC_RING_FILL,IDC_SLIDER_RING_FILL);

  ConfigSliderSetWidth(hDlg,IDC_SLIDER_SHARE_SIZE,IDC_SLIDER_RING_FILL);
  ConfigSliderSetText(hDlg,IDC_STATIC_SHARE_SIZE,IDC_SLIDER_SHARE_SIZE);
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
void ConfigOptionsSyncShareMode(HWND hDlg)
{
  int eShareMode;
  int autoconvertpcm;

  eShareMode=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE         // _In_  int nIDButton
  );

  eShareMode=eShareMode?AUDCLNT_SHAREMODE_EXCLUSIVE:AUDCLNT_SHAREMODE_SHARED;

  if (AUDCLNT_SHAREMODE_EXCLUSIVE==eShareMode) {
    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),
                                // _In_  HWND hWnd,
      FALSE                     // _In_  BOOL bEnable
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),
                                // _In_  HWND hWnd,
      FALSE                     // _In_  BOOL bEnable
    );
  }
  else {
    autoconvertpcm=IsDlgButtonChecked(
      hDlg,                     // _In_  HWND hDlg,
      IDC_CHECKBOX_AUTOCONVERT_PCM
                                // _In_  int nIDButton
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),
                                // _In_  HWND hWnd,
      TRUE                      // _In_  BOOL bEnable
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),
                                // _In_  HWND hWnd,
      autoconvertpcm           // _In_  BOOL bEnable
    );
  }
}
#else // } {
void ConfigOptionsSyncShareMode(HWND hDlg)
{
  int exclusive;
  int autoconvertpcm;

  exclusive=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE         // _In_  int nIDButton
  );

  if (exclusive) {
    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),
                                // _In_  HWND hWnd,
      FALSE                     // _In_  BOOL bEnable
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),
                                // _In_  HWND hWnd,
      FALSE                     // _In_  BOOL bEnable
    );
  }
  else {
    autoconvertpcm=IsDlgButtonChecked(
      hDlg,                     // _In_  HWND hDlg,
      IDC_CHECKBOX_AUTOCONVERT_PCM
                                // _In_  int nIDButton
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),
                                // _In_  HWND hWnd,
      TRUE                      // _In_  BOOL bEnable
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),
                                // _In_  HWND hWnd,
      autoconvertpcm           // _In_  BOOL bEnable
    );
  }
}
#endif // }

#if defined (YASAPI_CONFIG2) // {
void ConfigOptionsSetDeviceOptions(HWND hDlg, const Options *pOptions,
    int nDevice)
#else // } {
static void ConfigOptionsSetDeviceOptions(HWND hDlg, const Options *pDevice)
#endif // }
{
#if defined (YASAPI_CONFIG2) // {
  const DeviceOptions *pDevice=
      pOptions?pOptions->pDevices+nDevice
      :GetDefaultDeviceOptions();
#else // } {

  if (NULL==pDevice)
    pDevice=GetDefaultOptions();
#endif // }

#if defined (YASAPI_CONFIG2) // {
  // The device name is not part of the default device options.
  // Hence don't change it when default options are requested.
  if (pOptions) {
    SetWindowTextW(
      GetDlgItem(hDlg,IDC_STATIC_DEVICE_NAME),
                                // _In_     HWND    hWnd,
      pDevice->val_Device_FriendlyName.pwszVal
                                // _In_opt_ LPCTSTR lpString
    );
  }
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // sharemode
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  CheckRadioButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE,        // _In_  int nIDFirstButton,
    IDC_RADIO_SHARED,           // _In_  int nIDLastButton,
    AUDCLNT_SHAREMODE_EXCLUSIVE==pDevice->eShareMode
        ?IDC_RADIO_EXCLUSIVE
        :IDC_RADIO_SHARED       // _In_  int nIDCheckButton
  );
#else // } {
  CheckRadioButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE,        // _In_  int nIDFirstButton,
    IDC_RADIO_AUTOMATIC,        // _In_  int nIDLastButton,
    YASAPI_MODE_EXCLUSIVE==pDevice->mode?IDC_RADIO_EXCLUSIVE:
        YASAPI_MODE_SHARED==pDevice->mode?IDC_RADIO_SHARED:
        IDC_RADIO_AUTOMATIC     // _In_  int nIDCheckButton
  );
#endif // }

  // push/pull
  CheckRadioButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_PUSH,             // _In_  int nIDFirstButton,
    IDC_RADIO_PULL,             // _In_  int nIDLastButton,
    0==pDevice->pull
        ?IDC_RADIO_PUSH
        :IDC_RADIO_PULL         // _In_  int nIDCheckButton
  );

  // auto convert pcm
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_AUTOCONVERT_PCM,
                                // _In_  int nIDButton,
    pDevice->autoconvertpcm    // _In_  UINT uCheck
  );

  // src default quality
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_SRC_DEFAULT_QUALITY,
                                // _In_  int nIDButton,
    pDevice->src_default_quality
                                // _In_  UINT uCheck
  );

  ConfigOptionsSyncShareMode(hDlg);
}

#if defined (YASAPI_CONFIG2) // {
void ConfigOptionsSetDeviceBuffers(HWND hDlg, const Options *pOptions,
    int nDevice)
#else // } {
static void ConfigOptionsSetDeviceBuffers(HWND hDlg, const Options *pDevice)
#endif // }
{
#if defined (YASAPI_CONFIG2) // {
  const DeviceOptions *pDevice=
      pOptions?pOptions->pDevices+nDevice
      :GetDefaultDeviceOptions();
#else // } {

  if (NULL==pDevice)
    pDevice=GetDefaultOptions();
#endif // }

#if defined (YASAPI_CONFIG2) // {
  // The device name is not part of the default device options.
  // Hence don't change it when default options are requested.
  if (pOptions) {
    SetWindowTextW(
      GetDlgItem(hDlg,IDC_STATIC_DEVICE_NAME),
                                // _In_     HWND    hWnd,
      pDevice->val_Device_FriendlyName.pwszVal
                                // _In_opt_ LPCTSTR lpString
    );
  }
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // ring size
  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_RING_SIZE),
                                // _In_  HWND hWnd,
    TBM_SETRANGE,               // _In_  UINT Msg,
    FALSE,                      // _In_  WPARAM wParam (redraw flag),
    MAKELONG(CONFIG_SLIDER_MIN,CONFIG_SLIDER_MAX)
                                // _In_  LPARAM lParam
  );

  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_RING_SIZE),
                                // _In_  HWND hWnd,
    TBM_SETPOS,                 // _In_  UINT Msg,
    TRUE,                       // _In_  WPARAM wParam (redraw flag),
    pDevice->qRingSize*CONFIG_SLIDER_SCALE+0.5
                                // _In_  LPARAM lParam
  );

  ConfigSliderSetText(hDlg,IDC_STATIC_RING_SIZE,IDC_SLIDER_RING_SIZE);

  // ring fill
  ConfigSliderSetWidth(hDlg,IDC_SLIDER_RING_FILL,IDC_SLIDER_RING_SIZE);

  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_RING_FILL),
                                // _In_  HWND hWnd,
    TBM_SETPOS,                 // _In_  UINT Msg,
    TRUE,                       // _In_  WPARAM wParam (redraw flag),
    pDevice->qRingFill*CONFIG_SLIDER_SCALE+0.5
                                // _In_  LPARAM lParam
  );

  ConfigSliderSetText(hDlg,IDC_STATIC_RING_FILL,IDC_SLIDER_RING_FILL);

  // share size
  ConfigSliderSetWidth(hDlg,IDC_SLIDER_SHARE_SIZE,IDC_SLIDER_RING_FILL);

  SendMessage(
    GetDlgItem(hDlg,IDC_SLIDER_SHARE_SIZE),
                                // _In_  HWND hWnd,
    TBM_SETPOS,                 // _In_  UINT Msg,
    TRUE,                       // _In_  WPARAM wParam (redraw flag),
    pDevice->qShareSize*CONFIG_SLIDER_SCALE+0.5
                                // _In_  LPARAM lParam
  );

  ConfigSliderSetText(hDlg,IDC_STATIC_SHARE_SIZE,IDC_SLIDER_SHARE_SIZE);
}

void ConfigOptionsSetOptions(HWND hDlg, const Options *pOptions)
{
  const Options *pOptions1=pOptions;

  if (NULL==pOptions)
    pOptions=GetDefaultOptions();

  /////////////////////////////////////////////////////////////////////////////
  // volume control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_VOLUME,        // _In_  int nIDButton,
    pOptions->volume            // _In_  UINT uCheck
  );

#if defined (YASAPI_OPTION_WRITTEN_TIME) // {
  /////////////////////////////////////////////////////////////////////////////
  // writtentime audio control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_AUDIO,
                                // _In_  int nIDButton,
    pOptions->writtentime.audio // _In_  UINT uCheck
  );

  /////////////////////////////////////////////////////////////////////////////
  // writtentime video control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_VIDEO,
                                // _In_  int nIDButton,
    pOptions->writtentime.video // _In_  UINT uCheck
  );
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // mono2stereo control
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_MONO2STEREO,   // _In_  int nIDButton,
    pOptions->mono2stereo       // _In_  UINT uCheck
  );

#if defined (YASAPI_CONFIG2) // {
  // options must be set to different pages with different hDlg.
#else // }
  ConfigOptionsSetDeviceOptions(hDlg,pOptions1);
  ConfigOptionsSetDeviceBuffers(hDlg,pOptions1);
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // debug
#if defined (YASAPI_DEBUG) // {
  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_CONSOLE,       // _In_  int nIDButton,
    0==pOptions->debug          // _In_  UINT uCheck
  );
#else // } {
  EnableWindow(
    GetDlgItem(hDlg,IDC_STATIC_LABEL_CONSOLE),
                                // _In_  HWND hWnd,
    FALSE                       // _In_  BOOL bEnable
  );

  CheckDlgButton(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_CONSOLE,       // _In_  int nIDButton,
    TRUE                        // _In_  UINT uCheck
  );

  EnableWindow(
    GetDlgItem(hDlg,IDC_CHECKBOX_CONSOLE),
                                // _In_  HWND hWnd,
    FALSE                       // _In_  BOOL bEnable
  );
#endif // }
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_CONFIG2) // {
void ConfigOptionsGetDeviceOptions(HWND hDlg, Options *pOptions, int nDevice)
#else // } {
static void ConfigOptionsGetDeviceOptions(HWND hDlg, Options *pDevice)
#endif // }
{
#if defined (YASAPI_CONFIG2) // {
  DeviceOptions *pDevice=pOptions->pDevices+nDevice;
#endif // }
#if ! defined (YASAPI_AUDCLNT_SHAREMODE) // {
  int checked;
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // sharemode
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  pDevice->eShareMode=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE         // _In_  int nIDButton
  );

  pDevice->eShareMode=pDevice->eShareMode
    ?AUDCLNT_SHAREMODE_EXCLUSIVE
		:AUDCLNT_SHAREMODE_SHARED;
#else // } {
  checked=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_EXCLUSIVE         // _In_  int nIDButton
  );

  if (checked) {
    pDevice->mode=YASAPI_MODE_EXCLUSIVE;
    goto strategy;
  }

  checked=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_SHARED            // _In_  int nIDButton
  );

  if (checked) {
    pDevice->mode=YASAPI_MODE_SHARED;
    goto strategy;
  }

  checked=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_AUTOMATIC         // _In_  int nIDButton
  );

  if (checked) {
    pDevice->mode=YASAPI_MODE_AUTOMATIC;
    goto strategy;
  }
strategy:
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // push / pull
  pDevice->pull=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_RADIO_PULL              // _In_  int nIDButton
  );

  // autoconv pcm
  pDevice->autoconvertpcm=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_AUTOCONVERT_PCM
                                // _In_  int nIDButton
  );

  // src default quality
  pDevice->src_default_quality=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_SRC_DEFAULT_QUALITY
                                // _In_  int nIDButton
  );
}

#if defined (YASAPI_CONFIG2) // {
void ConfigOptionsGetDeviceBuffers(HWND hDlg, Options *pOptions, int nDevice)
#else // } {
static void ConfigOptionsGetDeviceBuffers(HWND hDlg, Options *pDevice)
#endif // }
{
#if defined (YASAPI_CONFIG2) // {
  DeviceOptions *pDevice=pOptions->pDevices+nDevice;
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  pDevice->qRingSize=ConfigSliderGet(hDlg,IDC_SLIDER_RING_SIZE);
  pDevice->qRingFill=ConfigSliderGet(hDlg,IDC_SLIDER_RING_FILL);
  pDevice->qShareSize=ConfigSliderGet(hDlg,IDC_SLIDER_SHARE_SIZE);
}

void ConfigOptionsGetOptions(HWND hDlg, Options *pOptions)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pOptions->debug;
#endif // }

#if defined (YASAPI_CONFIG2) // {
  // options must be set to different pages with different hDlg.
#else // }
  ConfigOptionsGetDeviceBuffers(hDlg,pOptions);
  ConfigOptionsGetDeviceOptions(hDlg,pOptions);
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // mono2stereo control
  pOptions->mono2stereo=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_MONO2STEREO    // _In_  int nIDButton
  );

#if defined (YASAPI_OPTION_WRITTEN_TIME) // {
  /////////////////////////////////////////////////////////////////////////////
  // writtentime video control
  pOptions->writtentime.video=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_VIDEO
                                // _In_  int nIDButton
  );

  /////////////////////////////////////////////////////////////////////////////
  // writtentime audio control
  pOptions->writtentime.audio=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_WRITTENTIME_AUDIO
                                // _In_  int nIDButton
  );
#endif // }

  /////////////////////////////////////////////////////////////////////////////
  // volume control
  pOptions->volume=IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_VOLUME         // _In_  int nIDButton
  );

#if defined (YASAPI_DEBUG) // {
  /////////////////////////////////////////////////////////////////////////////
  // debug
  pOptions->debug=FALSE==IsDlgButtonChecked(
    hDlg,                       // _In_  HWND hDlg,
    IDC_CHECKBOX_CONSOLE        // _In_  int nIDButton
  );

  OptionsSave(pOptions,debug);
#else // } {
  OptionsSave(pOptions);
#endif // }
}
