/*
 * yasapi_options.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

#define YASAPI_OPTION_FILE \
    L"Plugins\\" YASAPI_WSTR(YASAPI_ID) L".ini"

#define YASAPI_APP_PLUGIN             L"plugin"
#if ! defined (YASAPI_CONFIG2) // {
#define YASAPI_APP_WASAPI             L"wasapi"
#endif // }
#define YASAPI_KEY_MONO2STEREO        L"mono2stereo"
#define YASAPI_KEY_VOLUME             L"volume"
#if defined (YASAPI_OPTION_WRITTEN_TIME) // {
#define YASAPI_KEY_WRITTENTIME_AUDIO  L"writtentime_audio"
#define YASAPI_KEY_WRITTENTIME_VIDEO  L"writtentime_video"
#endif // }
#if defined (YASAPI_CONFIG2) // {
#define YASAPI_KEY_PAGE               L"page"
#define YASAPI_KEY_ID                 L"device"
#endif // }
#define YASAPI_KEY_DEBUG              L"debug"

int OptionsCreate(Options *pOptions)
{
  char *dir;
  size_t len1,len2;
  wchar_t *wp;

  ZeroMemory(
    pOptions,           // [in]  PVOID Destination,
    sizeof *pOptions    // [in]  SIZE_T Length
  );

  dir=(char *)SendMessageW(winampGetOutModule()->hMainWindow,
      WM_WA_IPC,0,IPC_GETINIDIRECTORY);

  if (NULL==dir)
    goto dir;

  len1=MultiByteToWideChar(
    CP_ACP,             // _In_      UINT   CodePage,
    0,                  // _In_      DWORD  dwFlags,
    dir,                // _In_      LPCSTR lpMultiByteStr,
    -1,                 // _In_      int    cbMultiByte,
    NULL,               // _Out_opt_ LPWSTR lpWideCharStr,
    0                   // _In_      int    cchWideChar
  );

  --len1;
  len2=wcslen(YASAPI_OPTION_FILE);

  pOptions->path=(wchar_t *)GlobalAlloc(
    GMEM_FIXED,         // _In_  UINT uFlags,
    ((len1+1)+(len2+1))*(sizeof *pOptions->path)
                        // _In_  SIZE_T dwBytes
  );

  if (NULL==pOptions->path)
    goto path;
  
  wp=pOptions->path;

  wp+=MultiByteToWideChar(
    CP_ACP,             // _In_      UINT   CodePage,
    0,                  // _In_      DWORD  dwFlags,
    dir,                // _In_      LPCSTR lpMultiByteStr,
    -1,                 // _In_      int    cbMultiByte,
    wp,                 // _Out_opt_ LPWSTR lpWideCharStr,
    len1+1              // _In_      int    cchWideChar
  );

  --wp;
  *wp++='\\';

  memcpy(wp,YASAPI_OPTION_FILE,len2*(sizeof *wp));
  wp+=len2;
  *wp++=0;

#if defined (YASAPI_CONFIG2) // {
  pOptions->nDevice=-1;
  pOptions->cDevices=0;
  pOptions->pDevices=NULL;
#endif // }

  OptionsLoad(pOptions);

  return 0;
// cleanup:
  GlobalFree(pOptions->path);
path:
dir:
  return -1;
}

void OptionsDestroy(Options *pOptions)
{
  GlobalFree(pOptions->path);
}

#if defined (YASAPI_CONFIG2) // {
void OptionsSetDevice(Options *pOptions, int nDevice)
{
  pOptions->nDevice=nDevice;
  wcscpy(pOptions->id,pOptions->pDevices[nDevice].id);
}
#endif // }

void OptionsLoad(Options *pOptions)
{
  enum { SIZE=32 };
  wchar_t set[SIZE];
  wchar_t get[SIZE];
#if defined (YASAPI_CONFIG2) // {
  int nDevice;
  DeviceOptions *pDevice;
#endif // }

#if defined (YASAPI_DEBUG) // {
  // get the debug flag.
  pOptions->debug=GetPrivateProfileIntW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_DEBUG,             // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->debug,   // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );
#endif // }

  // get the mono2stereo flag.
  pOptions->mono2stereo=GetPrivateProfileIntW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_MONO2STEREO,       // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->mono2stereo,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the volume flag.
  pOptions->volume=GetPrivateProfileIntW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_VOLUME,            // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->volume,  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

#if defined (YASAPI_OPTION_WRITTEN_TIME) // {
  // get the writtentime audio flag.
  pOptions->writtentime.audio=GetPrivateProfileIntW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_WRITTENTIME_AUDIO, // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->writtentime.audio,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the writtentime video flag.
  pOptions->writtentime.video=GetPrivateProfileIntW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_WRITTENTIME_VIDEO, // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->writtentime.video,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );
#endif // }

#if defined (YASAPI_CONFIG2) // {
  // get the page.
  pOptions->nPage=GetPrivateProfileIntW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_PAGE,              // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->nPage,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  GetPrivateProfileStringW(
    YASAPI_APP_PLUGIN,            // _In_   LPCTSTR lpAppName,
    YASAPI_KEY_ID,                // _In_   LPCTSTR lpKeyName,
    GetDefaultOptions()->id,      // _In_   LPCTSTR lpDefault,
    pOptions->id,                 // _Out_  LPTSTR lpReturnedString,
    YASAPI_ID_SIZE,               // _In_   DWORD nSize,
    pOptions->path                // _In_   LPCTSTR lpFileName
  );
#else // } {
  DeviceOptionsLoad(pOptions,YASAPI_APP_WASAPI,GetDefaultOptions);
#endif // }
}

#if defined (YASAPI_DEBUG) // {
void OptionsSave(Options *pOptions, int debug)
#else // } {
void OptionsSave(Options *pOptions)
#endif // }
{
  enum { SIZE=32 };
  wchar_t buf[SIZE];
#if defined (YASAPI_CONFIG2) // {
  DeviceOptions *pDevice=pOptions->pDevices+pOptions->nDevice;
#endif // }

#if defined (YASAPI_OPTION_WRITTEN_TIME) // {
  // write the written time video flag.
  swprintf(buf,SIZE,L"%d",pOptions->writtentime.video);

  WritePrivateProfileStringW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_WRITTENTIME_VIDEO, // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the written time audio flag.
  swprintf(buf,SIZE,L"%d",pOptions->writtentime.audio);

  WritePrivateProfileStringW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_WRITTENTIME_AUDIO, // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );
#endif // }

  // write the volume flag.
  swprintf(buf,SIZE,L"%d",pOptions->volume);

  WritePrivateProfileStringW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_VOLUME,            // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the mono2stereo flag.
  swprintf(buf,SIZE,L"%d",pOptions->mono2stereo);

  WritePrivateProfileStringW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_MONO2STEREO,       // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

#if defined (YASAPI_CONFIG2) // {
  // write the page.
  swprintf(buf,SIZE,L"%d",pOptions->nPage);

  WritePrivateProfileStringW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_PAGE,              // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the id.
  WritePrivateProfileStringW(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_ID,                // _In_  LPCTSTR lpKeyName,
    pOptions->id,                 // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  DeviceOptionsSave(pDevice,pDevice->id);
#else // } {
  DeviceOptionsSave(pOptions,YASAPI_APP_WASAPI);
#endif // }

#if defined (YASAPI_DEBUG) // {
    swprintf(buf,SIZE,L"%d",pOptions->debug);

    WritePrivateProfileStringW(
      YASAPI_APP_PLUGIN,          // _In_  LPCTSTR lpAppName,
      YASAPI_KEY_DEBUG,           // _In_  LPCTSTR lpKeyName,
      buf,                        // _In_  LPCTSTR lpString,
      pOptions->path              // _In_  LPCTSTR lpFileName
    );

  if (pOptions->debug!=debug) {
    if (0==pOptions->debug)
      ConsoleClose();
    else
      ConsoleOpen();
  }
#endif // }
}

#if 0 // {
void OptionsCopy(Options *lhs, const Options *rhs)
{
  char *path;

  path=lhs->path;
  *lhs=*rhs;
  lhs->path=path;
}
#endif // }

const Options *GetDefaultOptions(void)
{
  static int initialized=0;
  static Options options;

  if (0==initialized) {
#if defined (YASAPI_DEBUG) // {
    options.debug=TRUE;
#endif // }
    options.mono2stereo=TRUE;
    options.volume=FALSE;
#if defined (YASAPI_OPTION_WRITTEN_TIME) // {
    options.writtentime.audio=TRUE;
    options.writtentime.video=TRUE;
#endif // }
#if defined (YASAPI_CONFIG2) // {
    options.id[0]=L'\0';
#else // } {
    options.eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    options.pull=FALSE;
    options.autoconvertpcm=TRUE;
    options.src_default_qualtiy=FALSE;
    options.pull=FALSE;
    options.qRingSize=4.0;
    options.qRingFill=2.5;
    options.qShareSize=0.0;
#endif // }
    initialized=1;
  }

  return &options;
}

#if defined (YASAPI_CONFIG2) // {
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
AUDCLNT_SHAREMODE OptionsGetShareMode(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].eShareMode;
}
#else // } {
int OptionsGetMode(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].mode;
}
#endif // }

int OptionsGetPull(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].pull;
}

int OptionsGetAutoConvertPCM(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].autoconvertpcm;
}

int OptionsGetSRCDefaultQuality(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].src_default_quality;
}

double OptionsGetRingSize(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].qRingSize;
}

double OptionsGetRingFill(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].qRingFill;
}

double OptionsGetShareSize(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].qShareSize;
}

void OptionsSetDevicePeriod(const Options *pOptions,
    REFERENCE_TIME hnsDevicePeriod)
{
  DPRINTF(pOptions->debug,"  setting device period to %I64d\n",
      hnsDevicePeriod);

  pOptions->pDevices[pOptions->nDevice].hnsDevicePeriod=hnsDevicePeriod;
}

REFERENCE_TIME OptionsGetDevicePeriod(const Options *pOptions)
{
  return pOptions->pDevices[pOptions->nDevice].hnsDevicePeriod;
}
#else // } {
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
AUDCLNT_SHAREMODE OptionsGetShareMode(const Options *pOptions)
{
  return pOptions->eShareMode;
}
#else // } {
int OptionsGetMode(const Options *pOptions)
{
  return pOptions->mode;
}
#endif // }

int OptionsGetPull(const Options *pOptions)
{
  return pOptions->pull;
}

int OptionsGetAutoConvertPCM(const Options *pOptions)
{
  return pOptions->autoconvertpcm;
}

int OptionsGetSRCDefaultQuality(const Options *pOptions)
{
  return pOptions->src_default_quality;
}

double OptionsGetRingSize(const Options *pOptions)
{
  return pOptions->qRingSize;
}

double OptionsGetRingFill(const Options *pOptions)
{
  return pOptions->qRingFill;
}

double OptionsGetShareSize(const Options *pOptions)
{
  return pOptions->qShareSize;
}

void OptionsSetDevicePeriod(const Options *pOptions,
    REFERENCE_TIME hnsDevicePeriod)
{
  DPRINTF(pOptions->debug,"  setting device period to %I64d\n",
      hnsDevicePeriod);

  pOptions->hnsDevicePeriod=hnsDevicePeriod;
}

REFERENCE_TIME OptionsGetDevicePeriod(const Options *pOptions)
{
  return pOptions->hnsDevicePeriod;
}
#endif // }

#if defined (YASAPI_DEBUG) // {
void OptionsDebug(Options *pOptions)
{
  int debug=pOptions->debug;

  DWPRINTF(debug,L"  options: \"%s\"\n",pOptions->path);
#if defined (YASAPI_AUDCLNT_SHAREMODE) // {
  DPRINTF(debug,"    share mode: %s\n",
      YASAPI_SHAREMODE_LABEL(OptionsGetShareMode(pOptions)));
#else // } {
  DPRINTF(debug,"    mode: %s\n",
      YASAPI_MODE_LABEL(OptionsGetMode(pOptions)));
#endif // }
  DPRINTF(debug,"    startegy: %s\n",
      OptionsGetPull(pOptions)?"pull":"push");
  DPRINTF(debug,"    ring size: %.1f\n",
      OptionsGetRingSize(pOptions));
  DPRINTF(debug,"    ring filled: %.1f\n",
      OptionsGetRingFill(pOptions));
  DPRINTF(debug,"    shared size: %.1f\n",
      OptionsGetShareSize(pOptions));
}
#endif // }
